@@ -1,9 +1,7 @@
 from __future__ import annotations
 
-from collections import deque
 from contextlib import AsyncExitStack, ExitStack
 from typing import (
-    AbstractSet,
     Any,
     Deque,
     Dict,
@@ -11,16 +9,14 @@ from typing import (
     Mapping,
     NamedTuple,
     Optional,
-    Set,
     Tuple,
     Union,
 )
 
 from graphlib2 import TopologicalSorter
 
-from di._utils.scope_validation import validate_scopes
+from di._utils.scope_map import ScopeMap
 from di._utils.task import AsyncTask, ExecutionState, SyncTask, gather_new_tasks
-from di.api.dependencies import DependantBase
 from di.api.executor import State as ExecutorState
 from di.api.executor import Task as ExecutorTask
 from di.api.providers import DependencyProvider
@@ -33,7 +29,7 @@ Task = Union[AsyncTask, SyncTask]
 
 TaskCacheDeque = Deque[Tuple[Task, Scope]]
 
-Results = Dict[Task, Any]
+Results = Dict[int, Any]
 
 TaskDependencyCounts = Dict[Task, int]
 
@@ -43,25 +39,16 @@ class SolvedDependantCache(NamedTuple):
 
     root_task: Task
     topological_sorter: TopologicalSorter[Task]
-    cacheable_tasks: AbstractSet[Task]
-    cached_tasks: Iterable[Tuple[Task, Scope]]
-    validated_scopes: Set[Tuple[Scope, ...]]
-    call_map: Mapping[DependencyProvider, Iterable[Task]]
+    callable_to_task_mapping: Mapping[DependencyProvider, Iterable[Task]]
 
 
 def plan_execution(
     stacks: Mapping[Scope, Union[AsyncExitStack, ExitStack]],
-    cached_values: Mapping[DependantBase[Any], Any],
+    cache: ScopeMap[DependencyProvider, Any],
     solved: SolvedDependant[Any],
     *,
     values: Optional[Mapping[DependencyProvider, Any]] = None,
-) -> Tuple[
-    Dict[Task, Any],
-    Iterable[Optional[ExecutorTask]],
-    ExecutorState,
-    Iterable[Tuple[Task, Scope]],
-    Task,
-]:
+) -> Tuple[Dict[int, Any], Iterable[Optional[ExecutorTask]], ExecutorState, Task,]:
     """Re-use or create an ExecutionPlan"""
     # This function is a hot loop
     # It is run for every execution, and even with the cache it can be a bottleneck
@@ -71,27 +58,13 @@ def plan_execution(
 
     solved_dependency_cache: SolvedDependantCache = solved.container_cache
 
-    scopes = tuple(stacks.keys())
-    if scopes not in solved_dependency_cache.validated_scopes:
-        validate_scopes(scopes, solved.dag)
-        solved_dependency_cache.validated_scopes.add(scopes)
-
-    results: Results = {}
-    call_map = solved_dependency_cache.call_map
+    # populate resulsts with values
+    results: "Results" = {}
+    call_map = solved_dependency_cache.callable_to_task_mapping
     for call in user_values.keys() & call_map.keys():
         value = user_values[call]
-        for task_id in call_map[call]:
-            results[task_id] = value
-
-    to_cache: TaskCacheDeque = deque()
-    cacheable_tasks = solved_dependency_cache.cacheable_tasks
-    add_to_cache = to_cache.append
-
-    for task, scope in solved_dependency_cache.cached_tasks:
-        if task.dependant in cached_values:
-            results[task] = cached_values[task.dependant]
-        elif task in cacheable_tasks:
-            add_to_cache((task, scope))
+        for task in call_map[call]:
+            results[task.task_id] = value
 
     ts = solved_dependency_cache.topological_sorter.copy()
 
@@ -99,12 +72,12 @@ def plan_execution(
         stacks=stacks,
         results=results,
         toplogical_sorter=ts,
+        cache=cache,
     )
 
     return (
         results,
         gather_new_tasks(execution_state),
         ExecutorState(execution_state),
-        to_cache,
         solved_dependency_cache.root_task,
     )
