@@ -1,6 +1,8 @@
 from __future__ import annotations
 
 import contextvars
+import inspect
+import sys
 from collections import deque
 from contextlib import contextmanager
 from types import TracebackType
@@ -15,65 +17,65 @@ from typing import (
     List,
     Mapping,
     Optional,
+    Sequence,
     Set,
-    Tuple,
     Type,
     TypeVar,
     Union,
 )
 
+if sys.version_info < (3, 8):
+    from typing_extensions import Protocol
+else:
+    from typing import Protocol
+
 from graphlib2 import TopologicalSorter
 
 from di._utils.execution_planning import SolvedDependantCache, plan_execution
-from di._utils.inspect import is_async_gen_callable, is_coroutine_callable
-from di._utils.nullcontext import nullcontext
+from di._utils.inspect import get_type, is_async_gen_callable, is_coroutine_callable
+from di._utils.scope_validation import validate_scopes
 from di._utils.state import ContainerState
 from di._utils.task import AsyncTask, SyncTask
 from di._utils.topsort import topsort
 from di._utils.types import FusedContextManager
-from di.api.dependencies import DependantBase, DependencyParameter
-from di.api.executor import AsyncExecutor, SyncExecutor
-from di.api.providers import DependencyProvider, DependencyProviderType
+from di.api.dependencies import CacheKey, DependantBase, DependencyParameter
+from di.api.executor import AsyncExecutorProtocol, SyncExecutorProtocol
+from di.api.providers import DependencyProvider
 from di.api.scopes import Scope
 from di.api.solved import SolvedDependant
-from di.dependant import Dependant
-from di.exceptions import SolvingError, WiringError
-from di.executors import DefaultExecutor
+from di.exceptions import WiringError
+
+__all__ = ("BaseContainer", "Container")
+
 
 _Task = Union[AsyncTask, SyncTask]
 
 _DependantTaskDag = Dict[_Task, Set[_Task]]
 _DependantQueue = Deque[DependantBase[Any]]
-_ExecutionCM = FusedContextManager[None]
-_nullcontext = nullcontext(None)
 
 
 DependencyType = TypeVar("DependencyType")
 
 
-__all__ = ("BaseContainer", "Container")
+class RegisterHook(Protocol):
+    def __call__(
+        self, param: Optional[inspect.Parameter], dependant: DependantBase[Any]
+    ) -> Optional[DependantBase[Any]]:
+        ...
 
 
 class _ContainerCommon:
-    __slots__ = ("_executor", "_execution_scope", "_binds")
+    __slots__ = ("_scopes", "_register_hooks")
 
-    _executor: Union[SyncExecutor, AsyncExecutor]
-    _execution_scope: Scope
-    _binds: Dict[DependencyProvider, DependantBase[Any]]
+    _scopes: Sequence[Scope]
+    _register_hooks: List[RegisterHook]
 
     def __init__(
         self,
-        executor: Union[SyncExecutor, AsyncExecutor],
-        execution_scope: Scope,
-        binds: Optional[Dict[DependencyProvider, DependantBase[Any]]],
+        scopes: Sequence[Scope],
     ):
-        self._executor = executor
-        self._execution_scope = execution_scope
-        self._binds = binds or {}
-
-    @property
-    def binds(self) -> Mapping[DependencyProvider, DependantBase[Any]]:
-        return self._binds
+        self._scopes = list(scopes)
+        self._register_hooks = []
 
     @property
     def scopes(self) -> Collection[Scope]:
@@ -83,12 +85,9 @@ class _ContainerCommon:
     def _state(self) -> ContainerState:
         raise NotImplementedError
 
-    def bind(
+    def register(
         self,
-        provider: Union[
-            DependantBase[DependencyType], DependencyProviderType[DependencyType]
-        ],
-        dependency: DependencyProviderType[DependencyType],
+        hook: RegisterHook,
     ) -> ContextManager[None]:
         """Replace a dependency provider with a new one.
 
@@ -98,24 +97,40 @@ class _ContainerCommon:
         Binds are only identified by the identity of the callable and do not take into account
         the scope or any other data from the dependency they are replacing.
         """
-        previous_provider = self.binds.get(dependency, None)
 
-        if not isinstance(provider, DependantBase):
-            provider = Dependant(provider)
-
-        self._binds[dependency] = provider
+        self._register_hooks.append(hook)
 
         @contextmanager
         def unbind() -> Generator[None, None, None]:
             try:
                 yield
             finally:
-                self._binds.pop(dependency)
-                if previous_provider is not None:
-                    self._binds[dependency] = previous_provider
+                self._register_hooks.remove(hook)
 
         return unbind()
 
+    def register_by_type(
+        self,
+        provider: DependantBase[Any],
+        dependency: DependencyProvider,
+    ) -> ContextManager[None]:
+        def hook(
+            param: Optional[inspect.Parameter], dependant: DependantBase[Any]
+        ) -> Optional[DependantBase[Any]]:
+            if dependant.call is dependency:
+                return provider
+            if param is None:
+                return None
+            type_annotation_option = get_type(param)
+            if type_annotation_option is None:
+                return None
+            type_annotation = type_annotation_option.value
+            if type_annotation is dependency:
+                return provider
+            return None
+
+        return self.register(hook)
+
     def solve(
         self,
         dependency: DependantBase[DependencyType],
@@ -124,20 +139,17 @@ class _ContainerCommon:
 
         Returns a SolvedDependant that can be executed to get the dependency's value.
         """
-        # If the SolvedDependant itself is a bind, replace it's dependant
-        if dependency.call in self._binds:
-            dependency = self._binds[dependency.call]
+        # If the dependency itself is a bind, replace it
+        for hook in self._register_hooks:
+            match = hook(None, dependency)
+            if match:
+                dependency = match
 
-        # Mapping of already seen dependants to itself for hash based lookups
-        dep_registry: Dict[DependantBase[Any], DependantBase[Any]] = {}
+        dependants: Dict[CacheKey, DependantBase[Any]] = {}
         # DAG mapping dependants to their dependendencies
         dep_dag: Dict[DependantBase[Any], List[DependantBase[Any]]] = {}
-
         # The same DAG as above but including parameters (inspect.Parameter instances)
-        param_graph: Dict[
-            DependantBase[Any],
-            List[DependencyParameter],
-        ] = {}
+        param_graph: Dict[DependantBase[Any], List[DependencyParameter]] = {}
 
         def get_params(
             dep: DependantBase[Any],
@@ -150,13 +162,10 @@ class _ContainerCommon:
                     param = param._replace(
                         dependency=param.dependency.register_parameter(param.parameter)
                     )
-                if (
-                    param.dependency.call is not None
-                    and param.dependency.call in self._binds
-                ):
-                    param = param._replace(
-                        dependency=self._binds[param.dependency.call]
-                    )
+                for hook in self._register_hooks:
+                    match = hook(param.parameter, param.dependency)
+                    if match is not None:
+                        param = param._replace(dependency=match)
                 params[idx] = param
                 if param.parameter is not None:
                     if (
@@ -171,34 +180,22 @@ class _ContainerCommon:
                         )
             return params
 
-        def check_equivalent(dep: DependantBase[Any]) -> None:
-            if dep in dep_registry and dep.scope != dep_registry[dep].scope:
-                raise SolvingError(
-                    f"The dependencies {dep} and {dep_registry[dep]}"
-                    " have the same lookup (__hash__ and __eq__) but have different scopes"
-                    f" ({dep.scope} and {dep_registry[dep].scope} respectively)"
-                    " This is likely a mistake, but you can override this behavior by either:"
-                    "\n  1. Wrapping the function in another function or subclassing the class such"
-                    " that they now are considered different dependencies"
-                    "\n  2. Use a custom implemetnation of DependantBase that changes the meaning"
-                    " or computation of __hash__ and/or __eq__"
-                )
-
         # Do a DFS of the DAG checking constraints along the way
         q: _DependantQueue = deque((dependency,))
         while q:
             dep = q.popleft()
-            if dep in dep_registry:
-                check_equivalent(dep)
+            cache_key = dep.cache_key
+            if cache_key in dependants:
+                continue
             else:
-                dep_registry[dep] = dep
+                dependants[cache_key] = dep
                 params = get_params(dep)
                 param_graph[dep] = params
                 dep_dag[dep] = []
                 for param in params:
                     predecessor_dep = param.dependency
                     dep_dag[dep].append(predecessor_dep)
-                    if predecessor_dep not in dep_registry:
+                    if predecessor_dep not in dependants:
                         q.append(predecessor_dep)
         # Filter out any dependencies that do not have a call
         # These do not become tasks since they don't need to be computed
@@ -213,21 +210,21 @@ class _ContainerCommon:
         # with references to all of their children
         dep_topsort = topsort(
             {
-                dep: [p.dependency for p in computable_param_graph[dep]]
-                for dep in computable_param_graph
+                dep.cache_key: [p.dependency.cache_key for p in params]
+                for dep, params in computable_param_graph.items()
             }
         )
         # Create a seperate TopologicalSorter to hold the Tasks
         ts: TopologicalSorter[_Task] = TopologicalSorter()
         tasks = self._build_tasks(
             computable_param_graph,
-            (node for nodes in dep_topsort for node in nodes),
+            (dependants[key] for key_group in dep_topsort for key in key_group),
             ts,
         )
         ts.prepare()
         task_dependency_dag: _DependantTaskDag = {
-            tasks[dep]: {
-                tasks[predecessor_dep.dependency]
+            tasks[dep.cache_key]: {
+                tasks[predecessor_dep.dependency.cache_key]
                 for predecessor_dep in predecessor_deps
             }
             for dep, predecessor_deps in computable_param_graph.items()
@@ -238,23 +235,17 @@ class _ContainerCommon:
                 call_map[t.call] = set()
             call_map[t.call].add(t)
         container_cache = SolvedDependantCache(
-            root_task=tasks[dependency],
+            root_task=tasks[dependency.cache_key],
             topological_sorter=ts,
-            cacheable_tasks={
-                tasks[d]
-                for d in computable_param_graph
-                if d.scope != self._execution_scope and d.share
-            },
-            cached_tasks=tuple(
-                {(tasks[d], d.scope) for d in computable_param_graph if d.share}
-            ),
-            validated_scopes=set(),
-            call_map={k: tuple(v) for k, v in call_map.items()},
+            callable_to_task_mapping={k: tuple(v) for k, v in call_map.items()},
+        )
+        validate_scopes(
+            self._scopes,
+            dep_dag,
         )
         solved = SolvedDependant(
             dependency=dependency,
             dag=param_graph,
-            topsort=dep_topsort,
             container_cache=container_cache,
         )
         return solved
@@ -267,14 +258,15 @@ class _ContainerCommon:
         ],
         topsorted: Iterable[DependantBase[Any]],
         ts: TopologicalSorter[_Task],
-    ) -> Dict[DependantBase[Any], _Task]:
-        tasks: Dict[DependantBase[Any], _Task] = {}
+    ) -> Dict[CacheKey, _Task]:
+        tasks: Dict[CacheKey, _Task] = {}
+        task_id = 0
         for dep in topsorted:
             positional: List[_Task] = []
             keyword: Dict[str, _Task] = {}
             for param in dag[dep]:
                 if param.parameter is not None:
-                    task = tasks[param.dependency]
+                    task = tasks[param.dependency.cache_key]
                     if param.parameter.kind is param.parameter.KEYWORD_ONLY:
                         keyword[param.parameter.name] = task
                     else:
@@ -283,40 +275,36 @@ class _ContainerCommon:
             positional_parameters = tuple(positional)
             keyword_parameters = tuple((k, v) for k, v in keyword.items())
 
+            assert dep.call is not None
             if is_async_gen_callable(dep.call) or is_coroutine_callable(dep.call):
-                tasks[dep] = task = AsyncTask(
+                tasks[dep.cache_key] = task = AsyncTask(
+                    scope=dep.scope,
+                    call=dep.call,
+                    use_cache=dep.use_cache,
                     dependant=dep,
+                    task_id=task_id,
                     positional_parameters=positional_parameters,
                     keyword_parameters=keyword_parameters,
                 )
             else:
-                tasks[dep] = task = SyncTask(
+                tasks[dep.cache_key] = task = SyncTask(
+                    scope=dep.scope,
+                    call=dep.call,
+                    use_cache=dep.use_cache,
                     dependant=dep,
+                    task_id=task_id,
                     positional_parameters=positional_parameters,
                     keyword_parameters=keyword_parameters,
                 )
-            ts.add(task, *(tasks[p.dependency] for p in dag[dep]))
+            task_id += 1
+            ts.add(task, *(tasks[p.dependency.cache_key] for p in dag[dep]))
         return tasks
 
-    def _update_cache(
-        self,
-        state: ContainerState,
-        results: Dict[_Task, Any],
-        to_cache: Iterable[Tuple[_Task, Scope]],
-    ) -> None:
-        cache = state.cached_values.set
-        for task in to_cache:
-            cache(
-                task[0].dependant,
-                results[task[0]],
-                scope=task[1],
-            )
-
     def execute_sync(
         self,
         solved: SolvedDependant[DependencyType],
+        executor: SyncExecutorProtocol,
         *,
-        executor: Optional[SyncExecutor] = None,
         values: Optional[Mapping[DependencyProvider, Any]] = None,
     ) -> DependencyType:
         """Execute an already solved dependency.
@@ -324,51 +312,33 @@ class _ContainerCommon:
         This method is synchronous and uses a synchronous executor,
         but the executor may still be able to execute async dependencies.
         """
-        cm: _ExecutionCM
-        state = self._state
-        if self._execution_scope in state.stacks.keys():
-            cm = _nullcontext
-        else:
-            state = state.copy()
-            cm = state.enter_scope(self._execution_scope)
-        with cm:
-            results, leaf_tasks, execution_state, to_cache, root_task = plan_execution(
-                stacks=state.stacks,
-                cached_values=state.cached_values.to_mapping(),
-                solved=solved,
-                values=values,
-            )
-            if root_task not in results:
-                (executor or self._executor).execute_sync(leaf_tasks, execution_state)  # type: ignore[union-attr]
-                self._update_cache(state, results, to_cache)
-            return results[root_task]  # type: ignore[no-any-return]
+        results, leaf_tasks, execution_state, root_task = plan_execution(
+            stacks=self._state.stacks,
+            cache=self._state.cached_values,
+            solved=solved,
+            values=values,
+        )
+        if root_task.task_id not in results:
+            executor.execute_sync(leaf_tasks, execution_state)  # type: ignore[union-attr]
+        return results[root_task.task_id]  # type: ignore[no-any-return]
 
     async def execute_async(
         self,
         solved: SolvedDependant[DependencyType],
+        executor: AsyncExecutorProtocol,
         *,
-        executor: Optional[AsyncExecutor] = None,
         values: Optional[Mapping[DependencyProvider, Any]] = None,
     ) -> DependencyType:
         """Execute an already solved dependency."""
-        cm: _ExecutionCM
-        state = self._state
-        if self._execution_scope in state.stacks.keys():
-            cm = _nullcontext
-        else:
-            state = state.copy()
-            cm = state.enter_scope(self._execution_scope)
-        async with cm:
-            results, leaf_tasks, execution_state, to_cache, root_task = plan_execution(
-                stacks=state.stacks,
-                cached_values=state.cached_values.to_mapping(),
-                solved=solved,
-                values=values,
-            )
-            if root_task not in results:
-                await (executor or self._executor).execute_async(leaf_tasks, execution_state)  # type: ignore[union-attr]
-                self._update_cache(state, results, to_cache)
-            return results[root_task]  # type: ignore[no-any-return]
+        results, leaf_tasks, execution_state, root_task = plan_execution(
+            stacks=self._state.stacks,
+            cache=self._state.cached_values,
+            solved=solved,
+            values=values,
+        )
+        if root_task.task_id not in results:
+            await executor.execute_async(leaf_tasks, execution_state)  # type: ignore[union-attr]
+        return results[root_task.task_id]  # type: ignore[no-any-return]
 
 
 class BaseContainer(_ContainerCommon):
@@ -380,20 +350,13 @@ class BaseContainer(_ContainerCommon):
     def __init__(
         self,
         *,
-        execution_scope: Scope = None,
-        executor: Optional[Union[AsyncExecutor, SyncExecutor]] = None,
+        scopes: Sequence[Scope],
     ) -> None:
         super().__init__(
-            executor=executor or DefaultExecutor(),
-            execution_scope=execution_scope,
-            binds={},
+            scopes=scopes,
         )
         self.__state = ContainerState.initialize()
 
-    @property
-    def binds(self) -> Mapping[DependencyProvider, DependantBase[Any]]:
-        return self._binds
-
     @property
     def scopes(self) -> Collection[Scope]:
         return self.__state.stacks.keys()
@@ -404,11 +367,10 @@ class BaseContainer(_ContainerCommon):
 
     def copy(self: _BaseContainerType) -> _BaseContainerType:
         new = object.__new__(self.__class__)
-        new._execution_scope = self._execution_scope
-        new._executor = self._executor
-        # binds are shared
-        new._binds = self._binds
-        # cached values and scopes are not shared
+        new._scopes = self._scopes
+        # binds are use_cached
+        new._register_hooks = self._register_hooks
+        # cached values and scopes are not use_cached
         new.__state = self.__state.copy()
         return new  # type: ignore[no-any-return]
 
@@ -474,13 +436,10 @@ class Container(_ContainerCommon):
     def __init__(
         self,
         *,
-        execution_scope: Scope = None,
-        executor: Optional[Union[AsyncExecutor, SyncExecutor]] = None,
+        scopes: Sequence[Scope],
     ) -> None:
         super().__init__(
-            executor=executor or DefaultExecutor(),
-            execution_scope=execution_scope,
-            binds={},
+            scopes=scopes,
         )
         self._context = contextvars.ContextVar(f"{self}._context")
         self._context.set(ContainerState.initialize())
@@ -495,9 +454,8 @@ class Container(_ContainerCommon):
 
     def copy(self: _ContainerType) -> _ContainerType:
         new = object.__new__(self.__class__)
-        new._execution_scope = self._execution_scope
-        new._executor = self._executor
-        new._binds = self._binds
+        new._scopes = self._scopes
+        new._register_hooks = self._register_hooks
         new._context = self._context
         return new  # type: ignore[no-any-return]
 
