@@ -1,21 +1,10 @@
 from __future__ import annotations
 
 import inspect
-import sys
-from typing import Any, Iterable, List, Mapping, Optional, Type, TypeVar, cast, overload
+from typing import Any, Iterable, List, Mapping, Optional, TypeVar, overload
 
-if sys.version_info < (3, 8):
-    from typing_extensions import Protocol
-else:
-    from typing import Protocol
-
-if sys.version_info < (3, 9):
-    from typing_extensions import Annotated, get_args, get_origin
-else:
-    from typing import Annotated, get_args, get_origin
-
-from di._utils.inspect import get_parameters, infer_call_from_annotation
-from di.api.dependencies import DependantBase, DependencyParameter
+from di._utils.inspect import get_parameters, get_type
+from di.api.dependencies import CacheKey, DependantBase, DependencyParameter
 from di.api.providers import (
     AsyncGeneratorProvider,
     CallableProvider,
@@ -43,7 +32,7 @@ class Dependant(DependantBase[T]):
         call: Optional[AsyncGeneratorProvider[T]] = None,
         *,
         scope: Optional[Scope] = None,
-        share: bool = True,
+        use_cache: bool = True,
         wire: bool = True,
         overrides: Optional[Mapping[str, DependantBase[Any]]] = None,
         sync_to_thread: bool = False,
@@ -56,7 +45,7 @@ class Dependant(DependantBase[T]):
         call: Optional[CoroutineProvider[T]] = None,
         *,
         scope: Optional[Scope] = None,
-        share: bool = True,
+        use_cache: bool = True,
         wire: bool = True,
         overrides: Optional[Mapping[str, DependantBase[Any]]] = None,
         sync_to_thread: bool = False,
@@ -69,7 +58,7 @@ class Dependant(DependantBase[T]):
         call: Optional[GeneratorProvider[T]] = None,
         *,
         scope: Optional[Scope] = None,
-        share: bool = True,
+        use_cache: bool = True,
         wire: bool = True,
         overrides: Optional[Mapping[str, DependantBase[Any]]] = None,
         sync_to_thread: bool = False,
@@ -82,7 +71,7 @@ class Dependant(DependantBase[T]):
         call: Optional[CallableProvider[T]] = None,
         *,
         scope: Optional[Scope] = None,
-        share: bool = True,
+        use_cache: bool = True,
         wire: bool = True,
         overrides: Optional[Mapping[str, DependantBase[Any]]] = None,
         sync_to_thread: bool = False,
@@ -94,40 +83,25 @@ class Dependant(DependantBase[T]):
         call: Optional[DependencyProviderType[T]] = None,
         *,
         scope: Scope = None,
-        share: bool = True,
+        use_cache: bool = True,
         wire: bool = True,
         overrides: Optional[Mapping[str, DependantBase[Any]]] = None,
         sync_to_thread: bool = False,
     ) -> None:
         self.call = call
         self.scope = scope
-        self.share = share
+        self.use_cache = use_cache
         self.wire = wire
         self.overrides = overrides or {}
         self.sync_to_thread = sync_to_thread
 
-    def __hash__(self) -> int:
-        """Used to identify Dependants.
-        By default, we identify Dependant's by their callable.
-        See DependantBase for more details.
-        """
-        return id(self.call)
+    @property
+    def cache_key(self) -> CacheKey:
+        if self.use_cache is False or self.call is None:
+            return (self.__class__, id(self))
+        return (self.__class__, self.call, self.scope)
 
-    def __eq__(self, o: object) -> bool:
-        """Used to identify Dependants.
-        By default, just checks that both are marked as shared.
-        See DependantBase for more details.
-        """
-        if type(self) != type(o):
-            return False
-        o = cast(Dependant[T], o)
-        if self.share is False or o.share is False:
-            return False
-        return self.call is o.call and self.overrides == o.overrides
-
-    def register_parameter(
-        self: DependantBase[T], param: inspect.Parameter
-    ) -> DependantBase[T]:
+    def register_parameter(self, param: inspect.Parameter) -> DependantBase[Any]:
         """Hook to register the parameter this Dependant corresponds to.
 
         This can be used to inferr self.call from a type annotation (autowiring),
@@ -135,11 +109,12 @@ class Dependant(DependantBase[T]):
 
         This method can return the same or a new instance of a Dependant to avoid modifying itself.
         """
+        if self.wire is False:
+            return self
         if self.call is None:
-            if get_origin(param.annotation) is Annotated:
-                param = param.replace(annotation=next(iter(get_args(param.annotation))))
-            if param.annotation is not param.empty:
-                self.call = infer_call_from_annotation(param)
+            annotation_type_option = get_type(param)
+            if annotation_type_option is not None:
+                self.call = annotation_type_option.value
         return self
 
     def get_dependencies(self) -> List[DependencyParameter]:
@@ -155,36 +130,30 @@ class Dependant(DependantBase[T]):
                 continue
             else:
                 maybe_sub_dependant = next(get_markers_from_parameter(param), None)
-                if maybe_sub_dependant is None:
-                    sub_dependant = self.create_sub_dependant(param)
-                else:
+                if maybe_sub_dependant is not None:
                     sub_dependant = maybe_sub_dependant
+                elif param.default is param.empty:
+                    # try to auto-wire
+                    sub_dependant = Dependant[Any](
+                        call=None,
+                        scope=self.scope,
+                        use_cache=self.use_cache,
+                    )
+                else:
+                    # has a default parameter but we create a dependency anyway just for binds
+                    # but do not wire it to make autowiring less brittle and less magic
+                    sub_dependant = Dependant[Any](
+                        call=None,
+                        scope=self.scope,
+                        use_cache=self.use_cache,
+                        wire=False,
+                    )
             res.append(DependencyParameter(dependency=sub_dependant, parameter=param))
         return res
 
-    def create_sub_dependant(self, param: inspect.Parameter) -> DependantBase[Any]:
-        """Create a Dependant instance from a sub-dependency of this Dependency.
-
-        This is used in the scenario where a transient dependency is inferred from a type hint.
-        For example:
-
-        >>> class Foo:
-        >>>     ...
-        >>> def foo_factory(foo: Foo) -> Foo:
-        >>>     return foo
-        >>> def parent(foo: Annotated[..., Depends(foo_factory)]):
-        >>>    ...
-
-        In this scenario, `Depends(foo_factory)` will call
-        `Depends(foo_factory).create_sub_dependant(Parameter(Foo))`.
-
-        Usually you'll transfer `scope` and possibly `share`
-        to sub-dependencies created in this manner.
-        """
-        return Dependant[Any](
-            call=None,
-            scope=self.scope,
-            share=self.share,
+    def __repr__(self) -> str:
+        return (
+            f"{self.__class__.__name__}(call={self.call}, use_cache={self.use_cache})"
         )
 
 
@@ -203,7 +172,7 @@ class JoinedDependant(DependantBase[T]):
         self.dependant = dependant
         self.siblings = siblings
         self.scope = dependant.scope
-        self.share = dependant.share
+        self.use_cache = dependant.use_cache
 
     def get_dependencies(self) -> List[DependencyParameter]:
         """Get the dependencies of our main dependant and all siblings"""
@@ -212,50 +181,13 @@ class JoinedDependant(DependantBase[T]):
             *(DependencyParameter(dep, None) for dep in self.siblings),
         ]
 
-    def __hash__(self) -> int:
-        return hash((self.dependant, *self.siblings))
-
-    def __eq__(self, o: object) -> bool:
-        if not isinstance(o, JoinedDependant):
-            return False
-        return (self.dependant, *self.siblings) == (o.dependant, *o.siblings)
-
     def register_parameter(self, param: inspect.Parameter) -> DependantBase[T]:
         self.dependant = self.dependant.register_parameter(param)
         return self
 
+    @property
+    def cache_key(self) -> CacheKey:
+        return (self.dependant.cache_key, tuple((s.cache_key for s in self.siblings)))
 
-class CallableClass(Protocol[T]):
-    """A callable class that has a __call__ that is valid as dependency provider"""
-
-    __call__: DependencyProviderType[T]
-
-
-def CallableClassDependant(
-    call: Type[CallableClass[T]],
-    *,
-    instance_scope: Scope = None,
-    scope: Scope = None,
-    share: bool = True,
-    wire: bool = True,
-) -> Dependant[T]:
-    """Create a Dependant that will create and call a callable class
-
-    The class instance can come from the class' constructor (by default)
-    or be provided by cls_provider.
-    """
-    if not (inspect.isclass(call) and hasattr(call, "__call__")):
-        raise TypeError("call must be a callable class")
-    instance = Dependant[CallableClass[T]](
-        call,
-        scope=instance_scope,
-        share=True,
-        wire=wire,
-    )
-    return Dependant[T](
-        call=call.__call__,
-        scope=scope,
-        share=share,
-        wire=wire,
-        overrides={"self": instance},
-    )
+    def __repr__(self) -> str:
+        return f"{self.__class__.__name__}(dependant={self.dependant}, siblings={self.siblings})"
